/******************************************************
 * Sgorpi MultiTrack Writer
 * By: Hedde Bosman
 * 
 * This code is provided as is, without any warranty.
 * You are free to copy/use the code as long as you 
 * mention my name, in the code and/or about box.
 */
#include "WaveFile.h"


WaveFile::WaveFile(char *filename, bool rw, int smprate, waveType t, bool s) {
	WAVEFORMATEX wfx;

	f = NULL;

	readwrite = rw;

	f = fopen(filename, (rw?"wb":"rb") );
	if (f == NULL) {
		throw exception("Unable to open file");
	}
	if (rw) {
		stereo = s;
		wfx.nChannels = s ? 2 : 1;
		wfx.nSamplesPerSec = smprate;
		samplerate = smprate;
		type = t;
		switch (t) {
			case W08:
				wfx.wFormatTag = WAVE_FORMAT_PCM;
				wfx.wBitsPerSample = 8;
				wfx.nBlockAlign = wfx.nChannels;
				wfx.cbSize = 0;
				break;
			case W16:
				wfx.wFormatTag = WAVE_FORMAT_PCM;
				wfx.wBitsPerSample = 16;
				wfx.nBlockAlign = wfx.nChannels * 2;
				wfx.cbSize = 0;
				break;
			case W24:
				wfx.wFormatTag = WAVE_FORMAT_PCM;
				wfx.wBitsPerSample = 24;
				wfx.nBlockAlign = wfx.nChannels * 3;
				wfx.cbSize = 0;
				break;
			case W0_24:
				wfx.wFormatTag = WAVE_FORMAT_IEEE_FLOAT;
				wfx.wBitsPerSample = 32;
				wfx.nBlockAlign = wfx.nChannels * 4;
				wfx.cbSize = 0;
				break;
			case W24_0: // 24.0
				wfx.wFormatTag = WAVE_FORMAT_PCM;
				wfx.wBitsPerSample = 24;
				wfx.nBlockAlign = wfx.nChannels * 4;
				wfx.cbSize = 0;
				break;
			case W16_8:
			default: // 16.8
				wfx.wFormatTag = WAVE_FORMAT_PCM;
				wfx.wBitsPerSample = 32;
				wfx.nBlockAlign = wfx.nChannels * 4;
				wfx.cbSize = 2; // size of extra information following end of the WAVEFORMATEX structure. for some reason 0x0100...
				break;
		}
		wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
		// write wave start with fmt block start
		if (fwrite("RIFF    WAVEfmt ", 1, 16, f) != 16) {
			throw exception("wave header not written correctly");
		}
		filesize = 8; // don't count "RIFF####"
		// write fmt block SIZE with or without the cbsize argument depending on file type
		long test = sizeof(WAVEFORMATEX)-(type==W16_8?0:2);
		fwrite(&test, 1, sizeof(test), f);
		filesize += 4;

			// write fmt block with or without the cbsize argument depending on file type
		test = fwrite(&wfx, 1, sizeof(WAVEFORMATEX)-(type==W16_8?0:2), f);
		if (test != sizeof(WAVEFORMATEX)-(type==W16_8?0:2) ) {
			return;
		}
		filesize += sizeof(WAVEFORMATEX)-(type==W16_8?0:2);

		// write data block start 
		if (fwrite("data    ", 1, 8, f) != 8) {
			return;
		}
		filesize += 8;
		headersize = filesize;

	} else {/**************************************************************************/
		// read
		long waveformatSize;
		dataSize = 0;

		memset(hdr, 0, 5);
    

		fread(hdr, 1,4, f);
		if (strncmp(hdr, "RIFF",4) != 0) { throw exception("No RIFF header"); }

		fread(&filesize, sizeof(long), 1, f);

	    fread(hdr, 1,4, f);
		if (strncmp(hdr, "WAVE",4) != 0) { throw exception("No WAVE form-type"); }
    
		fread(hdr,1,4,f);
		if (strncmp(hdr, "fmt ",4) != 0) { throw exception("No wave-from chunk id"); }

		fread(&waveformatSize, sizeof(long), 1, f);
		fread(&wfx, sizeof(WAVEFORMATEX), 1, f);

	    printf("%s, fmt=%d, ch=%d smpr=%d, avbps=%d, bl=%d, bps=%d, cbsize=%d \n", hdr, 
			wfx.wFormatTag, wfx.nChannels, wfx.nSamplesPerSec, wfx.nAvgBytesPerSec,
			wfx.nBlockAlign, wfx.wBitsPerSample, wfx.cbSize);

		stereo = (wfx.nChannels == 2);
		samplerate = wfx.nSamplesPerSec;

		// type checking
		if (wfx.wBitsPerSample == 32) {
			if (wfx.wFormatTag == 1) { // 16.8
				type = W16_8;
			} else if (wfx.wFormatTag == 3) { // 0.24
				type = W0_24;
			}
		} else if (wfx.wBitsPerSample == 24) {
			if ( (wfx.nBlockAlign/wfx.nChannels)%4 == 0) { // 24.0
				type = W24_0;
			} else if ( (wfx.nBlockAlign/wfx.nChannels)%3 == 0) { // 24b
				type = W24;
			}
		} else if (wfx.wBitsPerSample == 16) {
			type = W16;
		} else if (wfx.wBitsPerSample == 8) {
			type = W08;
		}

		if (wfx.cbSize == 2) {
			fread(hdr, 1, 4, f);
		} else if (wfx.cbSize == 0) {
			fread(hdr, 1, 4, f);
		} else {
			hdr[0] = (char)(LOBYTE(wfx.cbSize));
			hdr[1] = (char)(HIBYTE(wfx.cbSize));
			fread(&hdr[2], 1, 2, f);
		}
		printf("hdr: %s\n", hdr);
		while (strncmp(hdr, "data",4) != 0) {
			fread(&waveformatSize, sizeof(long), 1, f);
			fseek(f, waveformatSize, SEEK_CUR);
			fread(hdr,1,4,f);
			printf("hdr: %s\n", hdr);
		}
		hdr[0] = hdr[0];
	}
}

WaveFile::~WaveFile() {
	if (f != NULL) {
		if (readwrite) {
			// write filesize set filesize-8 at position 4
			fseek(f, 4, SEEK_SET);
			fwrite(&filesize, sizeof(unsigned long), 1, f);
			// write datachunk size
			long tmp = filesize - headersize;
			fseek(f, headersize+4, SEEK_SET);
			fwrite(&tmp, sizeof(unsigned long), 1, f);

		}
		fclose(f);
	}
}

void WaveFile::checkReadDataSize() {
	if (dataSize <= 0 && !feof(f)) {
		fread(hdr, 1, 4, f);
		while (strncmp(hdr, "data",4) != 0 && !feof(f)) {
			fread(&dataSize, sizeof(long), 1, f);
			fseek(f, dataSize, SEEK_CUR);
			fread(hdr,1,4,f);
			printf("hdr: %s\n", hdr);
		}
		fread(&dataSize, sizeof(long), 1, f);
	}
}

// this function assumes that we have stopped reading after a header...
long WaveFile::readBuf(int numsmp, float *buf) {
	if (feof(f)) return 0;

	long i;
	int maxSmp = numsmp * (stereo? 2 : 1);
	float smp;
	char tmp[5];

	if (dataSize == 0 && strncmp(hdr, "data",4) == 0) {
		fread(&dataSize, sizeof(long), 1, f);
		if (feof(f)) return 0;
	}

	switch(type) {
		case W08:
			char csmp;
			for (i = 0; i < maxSmp && !feof(f); i++) {
				fread(&csmp, 1,1, f);
				buf[i] = csmp*128.0f;

				dataSize -= 1;
				checkReadDataSize();
			}
			break;
		case W16:
			short tmpsmp;
			for (i = 0; i < maxSmp && !feof(f); i++) {
				fread(&tmpsmp, 1,2, f);
				buf[i] = tmpsmp;

				dataSize -= 2;
				checkReadDataSize();
			}
			break;
		case W24:
			tmp[0]=tmp[1]=tmp[3] = 0;
			for (i = 0; i < maxSmp && !feof(f); i++) {
				fread(&tmp, 1,3, f);
				smp = (float) ((int)( (tmp[0]&0x80)<<24 | (tmp[0]&0x7F)<<16 | tmp[1]<<8 | tmp[2]));
				buf[i] = smp;

				dataSize -= 3;
				checkReadDataSize();
			}
			break;
		case W0_24:
			for (i = 0; i < maxSmp && !feof(f); i++) {
				fread(&smp, 4,1, f);
				buf[i] = smp*32768.0f;

				dataSize -= 4;
				checkReadDataSize();
			}
			break;
		case W24_0:
			for (i = 0; i < maxSmp && !feof(f); i++) {
				fread(&smp, 4,1, f);
				buf[i] = smp/256.0f;

				dataSize -= 4;
				checkReadDataSize();
			}
			break;
		case W16_8:
			for (i = 0; i < maxSmp && !feof(f); i++) {
				fread(&smp, 4,1, f);
				buf[i] = smp;

				dataSize -= 4;
				checkReadDataSize();
			}

			break;
	}
	return (stereo? i/2 : i);
}

void WaveFile::writeBuf(int numsmp, float *buf) {
	int i;
	int bufsize = numsmp * (stereo? 2 : 1);

	switch(type) {
		case W08: {
			char *samps = new char[bufsize];
			for (i = 0; i < bufsize; i++) {
				samps[i] = -128+(char) ( (buf[i]>32767.0f ? 32767: (buf[i]<-32768.0f ? -32768: buf[i]) ) / 256.0f);
			}
			if (fwrite(samps, sizeof(char), bufsize, f) != bufsize) {
			}
			filesize += bufsize*sizeof(char);
			delete [] samps;
			break;
		}
		case W16: {
			short *samps = new short[bufsize];
			for (i = 0; i < bufsize; i++) {
				samps[i] = (short) (buf[i]>32767.0f ? 32767: (buf[i]<-32768.0f ? -32768: buf[i]) );
				samps[i] = samps[i];
			}
			if (fwrite(samps, sizeof(short), bufsize, f) != bufsize) {
			}
			filesize += bufsize*sizeof(short);
			delete [] samps;
			break;
		}
		case W24: {
			char *samps = new char[bufsize*3];
			int smptmp;
			for (i = 0; i < bufsize; i++) {
				smptmp = (int)( (buf[i]>32767.0f ? 32767: (buf[i]<-32768.0f ? -32768: buf[i]) ) * 65536.0f);
				// fix
				samps[i*3+0] = smptmp >> 8;
				samps[i*3+1] = smptmp >> 16;
				samps[i*3+2] = smptmp >> 24;
			}
			if (fwrite(samps, 3, bufsize, f) != bufsize) {
			}
			filesize += bufsize*3;
			delete [] samps;
			break;
		}
		case W0_24: {// 0.24 waves are normalized from [-1,+1)
			float *samps = new float[bufsize];
			for (i = 0; i < bufsize; i++) {
				samps[i] = buf[i] / 32768.0f;
			}
			if (fwrite(samps, sizeof(float), bufsize, f) != bufsize) {
			}
			filesize += bufsize*sizeof(float);
			delete [] samps;
			break;
		}
		case W24_0: {// 24.0 waves are normalized from [-8milion,+8milion)
			float *samps = new float[bufsize];
			for (i = 0; i < bufsize; i++) {
				samps[i] = buf[i] * 256.0f;
			}
			if (fwrite(samps, sizeof(float), bufsize, f) != bufsize) {
			}
			filesize += bufsize*sizeof(float);
			delete [] samps;
			break;
		}
		case W16_8:
		default: // 16.8
			if (fwrite(buf, sizeof(float), bufsize, f) != bufsize) {
			}
			filesize += bufsize*sizeof(float);
			break;
	}

}

bool WaveFile::isStereo() { return stereo; }
int WaveFile::getSampleRate() { return samplerate; }
waveType WaveFile::getWaveType() { return type; }
