/******************************************************
 * Sgorpi MultiTrack Writer
 * By: Hedde Bosman
 * 
 * This code is provided as is, without any warranty.
 * You are free to copy/use the code as long as you 
 * mention my name, in the code and/or about box.
 */
#include "sgorpi_mtw.h"
#include "sgorpi_mtw_params.h"
#include "resrc1.h"

#ifdef _DEBUG
#define __DEBUG__ "d"
#else
#define __DEBUG__ ""
#endif

DLL_EXPORTS

//_Buzz code starts here
int numMachines = 0;
LookUp machines[64];
char outputDir[MAX_PATH]={0};

// constructor
mi::mi() {
	unsigned long tmp;

	GlobalVals = &gval;
	AttrVals = (int *)&aval;
//	TrackVals = tval;
	numTracks=0;

	ex.pmi = this;
	isStereo = false;
	doRecord = true;
	isRecording = false;
	firsttick = true;
	samplerate = 44100;
	updateVU = 2;
	maxValue = 0.0f;

	machines[numMachines].added = false;
	//** win32
	sgnObj = CreateEvent(NULL, false, false, NULL);
	outputFile = NULL;
	if (strlen(outputDir) ==0) { 
		GetCurrentDirectory(512, outputDir);
	}
	//
	machines[numMachines].thread = CreateThread(NULL, 0, &TrackWriter, this, 0, &tmp);
	SetThreadPriority(machines[numMachines].thread, THREAD_PRIORITY_BELOW_NORMAL);
	machines[numMachines++].pmi = this;
	wb_start = wb_end = NULL;

	wb_current = new WriteBlock;
	wb_current->next = NULL;
	writeBlockPos = 0;

	if (hwnd) {
		ShowWindow(hwnd, SW_HIDE);
	}
}


// destructor
mi::~mi() {
	// free all allocated buffers here
	int i = 0;
	for (i = 0; i < numMachines; i++) {
		if (machines[i].pmi == this) break;
	}
	DestroyWindow(machines[i].groupbox);
	DestroyWindow(machines[i].buttonRec);
	DestroyWindow(machines[i].sliderVol);
	DestroyWindow(machines[i].text);
	TerminateThread(machines[i].thread, 0);

	numMachines--;
	for (; i < numMachines; i++) {
		machines[i] = machines[i+1];
		setMixerColumnPos(i);
	}
	while(wb_start) {
		WriteBlock *tmp = wb_start;
		wb_start = wb_start->next;
		delete tmp;
	}
	if (hwnd) {
		ShowWindow(hwnd, SW_HIDE);
	}
	CloseHandle(sgnObj);
}


// prepare everything, first Tick and AttributesChanged called _after_ this function
void mi::MDKInit(CMachineDataInput * const pi) {
	deze = pCB->GetThisMachine();
	pCB->SetEventHandler(pCB->GetThisMachine(), DoubleClickMachine, (EVENT_HANDLER_PTR)&mi::HandleDoubleClick, NULL);
	if (pi) {
		pi->Read(doRecord);
		pi->Read(volume);
		pi->Read(outputDir, MAX_PATH);
		int i = WAV_0_24;
		pi->Read(i);
		wavetype = (WAVTYPE)i;
		gval.volume = (word) (volume*0xaaa9);
		gval.record = doRecord;
	} else {
		wavetype = WAV_0_24;
		volume = 1.0f;
		doRecord = true;
		gval.volume = (word) (volume*0xaaa9);
		gval.record = doRecord;
	}
}
void mi::MDKSave(CMachineDataOutput * const po) {
	if (po) {
		po->Write(doRecord);
		po->Write(volume);
		po->Write(outputDir, MAX_PATH);
		po->Write(wavetype);
	}
}
void mi::SetNumTracks(int const n) {
	numTracks=n;
}

// user has selected some menu options
void mi::Command(int const i) {
	if (i > 255) {
	} else {
		switch(i) {
			case 0:
				HandleDoubleClick(0);
				break;
			case 1:
				pCB->MessageBox("MultiTrack Writer   (built: "__DATE__ __DEBUG__")\nBy Sgorpi (sgorpi@sgorpi-music.com)\nhttp://www.sgorpi-music.com\n\nBig thanks to ALL buzz machine coders, including Polac, Cyanphase, BTDSys and Apo.");
				break;
			default: break;
		}
	}
}

// user has changed attributes, called after MDKInit()
void mi::AttributesChanged() {
	// check all the attributes in aval
	// aval.time // and so on...
}

// Describe the parameters
char const *mi::DescribeValue(int const param, int const value) {
	static char txt[50];
	txt[0]=0;

	switch(param) {
		case PARAM_VOL:
			sprintf(txt,"%3.3f%%", ((float)value)/((float)0xaaa9)*100);
			break;
		case PARAM_REC:
			sprintf(txt,"%s", (value?"Yes":"No"));
			break;
		case PARAM_OUT:
			sprintf(txt,"%s", (value?"Start":"Stop"));
			break;
		default:
			sprintf(txt,"%d, %d",value, param);
			break;
	}

	return txt;
}

// machine input has changed mono<->stereo
void mi::OutputModeChanged(bool stereo) {
	isStereo = stereo;
}

void mi::MidiNote(const int channel, const int value, const int velocity) {
	// velocity of '0' means note off...
}
void mi::Stop() {
	maxValue = -0.0f;
	StopRecording();
}

// check for updated parameters, called every tick
void mi::Tick() {
	bool updateDialog = false;
	if (firsttick) {
		firsttick = false;
		return;
	}
	if(pMasterInfo->SamplesPerSec != samplerate) {
		samplerate=pMasterInfo->SamplesPerSec;
		updateDialog = true;
	}

	if (gval.volume != paraVol.NoValue) {
		volume = ((float)gval.volume)/((float)0xaaa9);
		updateDialog = true;
	}
	if (gval.record != paraRec.NoValue) {
		doRecord = (gval.record == paraRec.MaxValue);
		updateDialog = true;
	}
	if (gval.output != paraOut.NoValue) {
		if (gval.output == paraOut.MaxValue) {
			StartRecording();
		} else {
			StopRecording();
		}
	}
//	for (int c = 0; c < numTracks; c++) {
//		tvals *tv=tval+c;
//		mytvals *mytv=mytval+c;
//	}

	if (updateDialog) {
		InvalidateRect(hwnd, NULL, false);
	}
}
void mi::StartRecording() {
	SendDlgItemMessage(hwnd, IDC_STATUS, WM_SETTEXT, 0, (LPARAM)"Recording...");
	for (int i = 0; i < numMachines; i++) {
		if (machines[i].pmi->doRecord && !machines[i].pmi->isRecording) {
			machines[i].pmi->isRecording = true;
			// open file function aanroep enzo
			WriteFileOpen(machines[i].pmi);
		}
	}
}

void mi::StopRecording() {
	for (int i = 0; i < numMachines; i++) {
		if (machines[i].pmi->isRecording) {
			machines[i].pmi->isRecording = false;
			// close file function aanroep enzo
			WriteFileClose(machines[i].pmi);
		}
	}
	if (hwnd) {
		PostMessage(hwnd, WM_USER+13, 0,0);
	}
}

bool mi::MDKWorkStereo(float *psamples, int numsamples, int const mode) {
	float vol = volume;
	if (mode==WM_NOIO) {
		// machine is not connected to anything
		return false;
	} else if (mode==WM_WRITE) {
		// no sound coming in
		//return false;
		vol = 0.0f;
	} else if (mode == WM_READ) {
		// <thru> set in the sequence editor
		//return true;
		vol = 1.0f;
	}

	// effects handle sound here usually
	// return true always, except false when you want to silence the output


	float tmp;
	updateVU--;
	if (updateVU <= 0) {
		MaxValueLeft = currentMaxValueLeft;
		MaxValueRight = currentMaxValueRight;
		currentMaxValueRight = currentMaxValueLeft = -390.0f;
	}
	if (isRecording) {
		do {
			psamples[0]=psamples[0]*vol;
			// keep track of maximal value;
			tmp = psamples[0]>0?psamples[0]:0-psamples[0];
			if (tmp > currentMaxValueLeft) { currentMaxValueLeft = tmp; }

			psamples[1]=psamples[1]*vol;
			// keep track of maximal value;
			tmp = psamples[1]>0?psamples[1]:0-psamples[1];
			if (tmp > currentMaxValueRight) { currentMaxValueRight = tmp; }

			wb_current->samples[writeBlockPos++] = psamples[0];
			wb_current->samples[writeBlockPos++] = psamples[1];

			if (writeBlockPos >= WRITEBLOCKSIZE) {
				wb_current->numSamples = WRITEBLOCKSIZE;
				if (wb_end) {
					wb_end->next = wb_current;
					wb_end = wb_current;
				} else {
					wb_start = wb_end = wb_current;
				}
				wb_current = new WriteBlock;
				wb_current->next = NULL;
				writeBlockPos = 0;
				SetEvent(sgnObj); // signal writer thread
			}

			psamples+=2;
		} while(--numsamples);
	} else {
		do {
			psamples[0]=psamples[0]*vol;
			// keep track of maximal value;
			tmp = psamples[0]>0?psamples[0]:0-psamples[0];
			if (tmp > currentMaxValueLeft) { currentMaxValueLeft = tmp; }

			psamples[1]=psamples[1]*vol;
			// keep track of maximal value;
			tmp = psamples[1]>0?psamples[1]:0-psamples[1];
			if (tmp > currentMaxValueRight) { currentMaxValueRight = tmp; }

			psamples+=2;
		} while(--numsamples);
	}
	if (maxValue > 5.0) {
		maxValue -= 5.0f;
	}
	if (currentMaxValueLeft > maxValue) maxValue = currentMaxValueLeft;
	if (currentMaxValueRight > maxValue) maxValue = currentMaxValueRight;

	if (hwnd && updateVU <= 0) {
		InvalidateRect(hwnd, &invalidRect, false);
		updateVU = UPDATEFREQ;
	}
	if (mode==WM_WRITE) { return false; }
	return true;
}

bool mi::MDKWork(float *psamples, int numsamples, int const mode) {
	float vol = volume;
	if (mode==WM_NOIO) {
		// machine is not connected to anything
		return false;
	} else if (mode==WM_WRITE) {
		// no sound coming in
		//return false;
		vol = 0.0f;
	} else if (mode == WM_READ) {
		// <thru> set in the sequence editor
		//return true;
		vol = 1.0f;
	}
	// effects handle sound here usually
	// return true always, except false when you want to silence the output

	float tmp;
	updateVU--;
	if (updateVU <= 0) {
		MaxValueLeft = currentMaxValueLeft;
		MaxValueRight = currentMaxValueRight;
		currentMaxValueRight = currentMaxValueLeft = -390.0f;
	}
	if (isRecording) {
		do {
			psamples[0]=psamples[0]*vol;
			// keep track of maximal value;
			tmp = psamples[0]>0?psamples[0]:0-psamples[0];
			if (tmp > currentMaxValueLeft) { currentMaxValueLeft = tmp; }
			
			wb_current->samples[writeBlockPos++] = psamples[0];

			if (writeBlockPos >= WRITEBLOCKSIZE) {
				wb_current->numSamples = WRITEBLOCKSIZE;
				// mutex start
				if (wb_end) {
					wb_end->next = wb_current;
					wb_end = wb_current;
				} else {
					wb_start = wb_end = wb_current;
				}
				// mutex stop

				wb_current = new WriteBlock;
				wb_current->next = NULL;
				writeBlockPos = 0;
				SetEvent(sgnObj); // signal writer thread
			}

			psamples++;
		} while(--numsamples);

	} else {
		do {
			psamples[0]=psamples[0]*vol;
			// keep track of maximal value;
			tmp = psamples[0]>0?psamples[0]:0-psamples[0];
			if (tmp > currentMaxValueLeft) { currentMaxValueLeft = tmp; }
			psamples++;
		} while(--numsamples);
	}
	if (maxValue > 5.0) {
		maxValue -= 5.0f;
	}
	currentMaxValueRight = currentMaxValueLeft;
	if (currentMaxValueLeft > maxValue) maxValue = currentMaxValueLeft;

	if (hwnd && updateVU <= 0) {
		InvalidateRect(hwnd, &invalidRect, false);
		updateVU = UPDATEFREQ;
	}
	return true;
}
