/******************************************************
 * Sgorpi MultiTrack Writer
 * By: Hedde Bosman
 * 
 * This code is provided as is, without any warranty.
 * You are free to copy/use the code as long as you 
 * mention my name, in the code and/or about box.
 */

/******************************************************
 * The idea of this plugin is that each class instance
 * gets some input and, when recording, writes it to a 
 * buffer.
 * Another thread then read that buffer and writes it 
 * to disk. That way sound playback/rendering and such 
 * hopefully won't hang waiting on the harddisk.
 */

#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include "../mdk.h"

// Machine settings

#define miMAX_TRACKS		0
/*
CMachineParameter const paraWord = { pt_word, "Desc","Long desc",0,0xfffe,0xffff,MPF_STATE,0};
CMachineParameter const paraByte = { pt_byte, "Desc","Long",0,0xfe,0xff,MPF_STATE,0};
CMachineParameter const paraSwitch = { pt_switch, "Desc","Long",SWITCH_OFF,SWITCH_ON,SWITCH_NO,MPF_STATE,SWITCH_OFF};
CMachineParameter const paraNote = { pt_note, "Desc","Long",NOTE_MIN,NOTE_OFF,NOTE_NO,MPF_STATE,10 };
*/

#pragma pack(1)

// storage for global parameters, updated every Tick()
class gvals {
public:
	word volume;		// word		// Example
	byte record;
	byte output;
//	byte byteparam;		// byte
//	byte switchparam;	// switch
//	byte noteparam;		// note
};

// storage for track parameters, updated every Tick()
class tvals {
public:
	byte cmd;
};

// storage for track parameters, updated after MDKInit() and AttributesChanged
class avals {
public:
	int midiIn;
};

#pragma pack()

class mi;

/**********************
 * other structures
 */
// must be even.
#define WRITEBLOCKSIZE 32768
#define UPDATEFREQ 20
struct WriteBlock {
	public:
		float samples[WRITEBLOCKSIZE];
		unsigned int numSamples;
		WriteBlock *next;
};
struct LookUp {
	public:
		mi *pmi;
		bool added;
		HWND buttonRec;
		HWND sliderVol;
		HWND groupbox;
		HWND text;
		HANDLE thread;
};
enum WAVTYPE {
	WAV_16=0, WAV_24, WAV_16_8, WAV_24_0, WAV_0_24, WAV_8, NONE,
};

extern int numMachines;
extern LookUp machines[64];


/*************************/

class miex : public CMDKMachineInterfaceEx {
	public:
		mi *pmi;
};

class mi : public CMDKMachineInterface {
public:
	mi();
	~mi();

// Buzz calls these functions
	void Command(int const i);
	void Tick();
	char const *DescribeValue(int const param, int const value);
	void MDKInit(CMachineDataInput * const pi);
	bool MDKWork(float *psamples, int numsamples, int const mode);
	bool MDKWorkStereo(float *psamples, int numsamples, int const mode);
	void MDKSave(CMachineDataOutput * const po);
	CMDKMachineInterfaceEx *GetEx() { return &ex; }
	void OutputModeChanged(bool stereo);
	void AttributesChanged();
	void MidiNote(int const channel, int const value, int const velocity);
	void Stop();

	miex ex;
	gvals gval;
	avals aval;
	void SetNumTracks(int const n);
//	tvals tval[miMAX_TRACKS];

	int numTracks;
	bool HandleDoubleClick(void * ptr);
	bool firsttick;

	float volume;
	bool isStereo;
	bool doRecord;
	bool isRecording;
	int samplerate;

	int updateVU;
	float currentMaxValueLeft;
	float currentMaxValueRight;
	float MaxValueLeft;
	float MaxValueRight;
	float maxValue;
	void StartRecording();
	void StopRecording();

	HANDLE sgnObj;
	FILE *outputFile;
	unsigned long outputFileWritten;
	unsigned long outputFileHeaderSz;

	CMachine *deze;
	WriteBlock *wb_start;
	WriteBlock *wb_end;
	WriteBlock *wb_current;
	int writeBlockPos;
private:
	// insert your machine's global variables here

//	struct mytvals {
//		byte note;
//	} mytval[miMAX_TRACKS];

};

extern HINSTANCE dllInstance;
extern HWND hwnd;
// gui stuff
extern BOOL WINAPI DllMain (HANDLE hModule, DWORD fwdreason, LPVOID lpReserved);
extern void setMixerColumnPos(int i);
extern RECT invalidRect;

extern char outputDir[MAX_PATH];
extern WAVTYPE wavetype;
// writer stuff
extern DWORD WINAPI TrackWriter(LPVOID lpParam);
extern void WriteFileOpen(mi *pmi);
extern void WriteFileClose(mi *pmi);
