/******************************************************
 * Sgorpi MultiTrack Writer
 * By: Hedde Bosman
 * 
 * This code is provided as is, without any warranty.
 * You are free to copy/use the code as long as you 
 * mention my name, in the code and/or about box.
 */
#include "sgorpi_mtw.h"
#include "resrc1.h"
#include <mmreg.h>
#include <time.h>

WAVTYPE wavetype = WAV_0_24;


void WriteFileOpen(mi *pmi) {
	WAVEFORMATEX wfx;
	char filename[MAX_PATH];
	unsigned long test;

	wfx.nChannels = pmi->isStereo? 2 : 1;
	wfx.nSamplesPerSec = pmi->samplerate;
	switch (wavetype) {
		case WAV_8:
			wfx.wFormatTag = WAVE_FORMAT_PCM;
			wfx.wBitsPerSample = 8;
			wfx.nBlockAlign = wfx.nChannels;
			wfx.cbSize = 0;
			break;
		case WAV_16:
			wfx.wFormatTag = WAVE_FORMAT_PCM;
			wfx.wBitsPerSample = 16;
			wfx.nBlockAlign = wfx.nChannels * 2;
			wfx.cbSize = 0;
			break;
		case WAV_24:
			wfx.wFormatTag = WAVE_FORMAT_PCM;
			wfx.wBitsPerSample = 24;
			wfx.nBlockAlign = wfx.nChannels * 3;
			wfx.cbSize = 0;
			break;
		case WAV_0_24:
			wfx.wFormatTag = WAVE_FORMAT_IEEE_FLOAT;
			wfx.wBitsPerSample = 32;
			wfx.nBlockAlign = wfx.nChannels * 4;
			wfx.cbSize = 0;
			break;
		case WAV_24_0: // 24.0
			wfx.wFormatTag = WAVE_FORMAT_PCM;
			wfx.wBitsPerSample = 24;
			wfx.nBlockAlign = wfx.nChannels * 4;
			wfx.cbSize = 0;
			break;
		case WAV_16_8:
		default: // 16.8
			wfx.wFormatTag = WAVE_FORMAT_PCM;
			wfx.wBitsPerSample = 32;
			wfx.nBlockAlign = wfx.nChannels * 4;
			wfx.cbSize = 2; // size of extra information following end of the WAVEFORMATEX structure. for some reason 0x0100...
			break;
	}
	wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
	// get the time
	time_t tm;
	time(&tm);
	// create filename with somewith unique name due to time
	sprintf(filename, "%s\\%s %d.wav", outputDir, pmi->pCB->GetMachineName(pmi->deze), ((int)tm)%86400);

	pmi->outputFile = fopen(filename, "wb");
	if (!pmi->outputFile) {
		MessageBox(NULL, "could not open output file", "error", MB_OK);
		pmi->doRecord = false;
		pmi->StopRecording();
		InvalidateRect(hwnd, NULL, false);
		return;
	}
	// write wave start with fmt block start
	if (fwrite("RIFF    WAVEfmt ", 1, 16, pmi->outputFile) != 16) {
		MessageBox(NULL, "wave start not written correctly", "error", MB_OK);
		pmi->doRecord = false;
		return;
	}
	pmi->outputFileWritten = 8; // don't count "RIFF####"
	// write fmt block SIZE with or without the cbsize argument depending on file type
	test = sizeof(WAVEFORMATEX)-(wavetype==WAV_16_8?0:2);
	fwrite(&test, 1, sizeof(test), pmi->outputFile);
	pmi->outputFileWritten += 4;

	// write fmt block with or without the cbsize argument depending on file type
	test = fwrite(&wfx, 1, sizeof(WAVEFORMATEX)-(wavetype==WAV_16_8?0:2), pmi->outputFile);
	if (test != sizeof(WAVEFORMATEX)-(wavetype==WAV_16_8?0:2) ) {
		MessageBox(NULL, "waveformatex not written correctly", "error", MB_OK);
		pmi->doRecord = false;
		return;
	}
	pmi->outputFileWritten += sizeof(WAVEFORMATEX)-(wavetype==WAV_16_8?0:2);
	// write data block start 
	if (fwrite("data    ", 1, 8, pmi->outputFile) != 8) {
		MessageBox(NULL, "data header not written correctly", "error", MB_OK);
		pmi->doRecord = false;
		return;
	}
	pmi->outputFileWritten += 8;
	pmi->outputFileHeaderSz = pmi->outputFileWritten;
}


void WriteFileBlock(mi *pmi, WriteBlock *b) {
	if (!pmi->outputFile) return;

	switch(wavetype) {
		unsigned int i;
		case WAV_8: {
			char samps[WRITEBLOCKSIZE];
			for (i = 0; i < b->numSamples; i++) {
				samps[i] = -128+(char) ( (b->samples[i]>32767.0f ? 32767: (b->samples[i]<-32768.0f ? -32768: b->samples[i]) ) / 256.0f);
			}
			if (fwrite(samps, sizeof(char), b->numSamples, pmi->outputFile) != b->numSamples) {
				MessageBox(NULL, "no. bytes not written correctly", "error", MB_OK);
			}
			pmi->outputFileWritten += b->numSamples*sizeof(char);
			break;
		}
		case WAV_16: {
			short samps[WRITEBLOCKSIZE];
			for (i = 0; i < b->numSamples; i++) {
				samps[i] = (short) (b->samples[i]>32767.0f ? 32767: (b->samples[i]<-32768.0f ? -32768: b->samples[i]) );
			}
			if (fwrite(samps, sizeof(short), b->numSamples, pmi->outputFile) != b->numSamples) {
				MessageBox(NULL, "no. bytes not written correctly", "error", MB_OK);
			}
			pmi->outputFileWritten += b->numSamples*sizeof(short);
			break;
		}
		case WAV_24: {
			char samps[WRITEBLOCKSIZE*3];
			int smptmp;
			for (i = 0; i < b->numSamples; i++) {
				smptmp = (int)( (b->samples[i]>32767.0f ? 32767: (b->samples[i]<-32768.0f ? -32768: b->samples[i]) ) * 65536.0f);
				// fix
				samps[i*3+0] = smptmp >> 8;
				samps[i*3+1] = smptmp >> 16;
				samps[i*3+2] = smptmp >> 24;
			}
			if (fwrite(samps, 3, b->numSamples, pmi->outputFile) != b->numSamples) {
				MessageBox(NULL, "no. bytes not written correctly", "error", MB_OK);
			}
			pmi->outputFileWritten += b->numSamples*3;
			break;
		}
		case WAV_0_24: // 0.24 waves are normalized from [-1,+1)
			for (i = 0; i < b->numSamples; i++) {
				b->samples[i] /= 32768.0f;
			}
			if (fwrite(b->samples, sizeof(float), b->numSamples, pmi->outputFile) != b->numSamples) {
				MessageBox(NULL, "no. bytes not written correctly", "error", MB_OK);
			}
			pmi->outputFileWritten += b->numSamples*sizeof(float);
			break;
		case WAV_24_0: // 24.0 waves are normalized from [-8milion,+8milion)
			for (i = 0; i < b->numSamples; i++) {
				b->samples[i] *= 256.0f;
			}
			if (fwrite(b->samples, sizeof(float), b->numSamples, pmi->outputFile) != b->numSamples) {
				MessageBox(NULL, "no. bytes not written correctly", "error", MB_OK);
			}
			pmi->outputFileWritten += b->numSamples*sizeof(float);
			break;
		case WAV_16_8:
		default: // 16.8
			// internal structure of buzz if i'm correct
			if (fwrite(b->samples, sizeof(float), b->numSamples, pmi->outputFile) != b->numSamples) {
				MessageBox(NULL, "no. bytes not written correctly", "error", MB_OK);
			}
			pmi->outputFileWritten += b->numSamples*sizeof(float);
			break;
	}
}


void WriteFileClose(mi *pmi) {
	unsigned long tmp;
	// maybe fixes bug?
	if (!pmi->outputFile) return;

	while(pmi->wb_start) {
		WriteBlock *tmp = pmi->wb_start;
		pmi->wb_start = pmi->wb_start->next;
		// write out tmp
		WriteFileBlock(pmi, tmp);
		// done
		delete tmp;
	}
	// maybe? mutex start
	pmi->wb_end = pmi->wb_start;

	// save last block
	if (pmi->wb_current != NULL) {
		pmi->wb_current->numSamples = pmi->writeBlockPos;
		WriteFileBlock(pmi, pmi->wb_current);
		delete pmi->wb_current;
	}
	pmi->wb_current = new WriteBlock;
	pmi->wb_current->next = NULL;
	pmi->writeBlockPos = 0;
	// check if file is actually open
	if (!pmi->outputFile) return;

	// write filesize set filesize-8 at position 4
	fseek(pmi->outputFile, 4, SEEK_SET);
	fwrite(&pmi->outputFileWritten, sizeof(unsigned long), 1, pmi->outputFile);
	// write datachunk size
	tmp = pmi->outputFileWritten - pmi->outputFileHeaderSz;
	fseek(pmi->outputFile, pmi->outputFileHeaderSz+4, SEEK_SET);
	fwrite(&tmp, sizeof(unsigned long), 1, pmi->outputFile);
	fclose(pmi->outputFile);

	pmi->outputFile = NULL;
}

// thread loop
DWORD WINAPI TrackWriter(LPVOID lpParam) { 
	mi *pmi = (mi*)lpParam;

	while(true) {
		WaitForSingleObject(pmi->sgnObj, INFINITE);

		while(pmi->wb_start) {
			WriteBlock *tmp = pmi->wb_start;
			pmi->wb_start = pmi->wb_start->next;
			// write out tmp
			WriteFileBlock(pmi, tmp);
			// done
			delete tmp;
		}
		// maybe? mutex start
		pmi->wb_end = pmi->wb_start;
		// maybe? mutex stop
	}
    return 0; 
}
